"""Various exceptions for the Awair API."""

from typing import Optional


class AwairError(Exception):
    """Base awair exception class."""

    message = "Error querying the Awair API."

    def __init__(self, extra_message: Optional[str] = None) -> None:
        """Add extra messages to our base message."""
        final_message = self.message
        if extra_message:
            final_message += f" {extra_message}"

        super().__init__(final_message)


class AuthError(AwairError):
    """Some kind of authorization or authentication failure."""

    message = (
        "The supplied access token is invalid or "
        + "does not have access to the requested data"
    )


class QueryError(AwairError):
    """The query was somehow malformed."""

    message = "The supplied parameters were invalid."


class NotFoundError(AwairError):
    """The requested endpoint is gone."""

    message = "The Awair API returned an unexpected HTTP 404."


class RatelimitError(AwairError):
    """The API quota was exceeded."""

    message = (
        "The ratelimit for the Awair API has been " + "exceeded. Please try again later"
    )



"""Sensors dict with attribute-like access."""

from python_awair.attrdict import AttrDict


class Sensors(AttrDict):
    """Sensors of an AwairDevice.

    A Sensors object represents a set of sensors and corresponding
    values for a given Awair device, at a given point in time. The
    object itself essentially inherits from *dict*, and thus one
    can access sensor values by their string keys, and it is iterable -
    just like a dict.

    However, the class also supports getting, setting, and deleting
    sensor values via dot-notation; like an attribute or property.

    For example, given a *foo* Sensors object with a "bar" sensor, you could
    access that value either by calling *foo["bar"]* or *foo.bar*.

    The sensor names from the Awair API are not entirely user-friendly,
    so we've aliased known sensors to more descriptive values:

    .. table::

        ======== ==========================
        API name python_awair name
        ======== ==========================
        temp     temperature
        humid    humidity
        co2      carbon_dioxide
        voc      volatile_organic_compounds
        pm25     particulate_matter_2_5
        lux      illuminance
        spl_a    sound_pressure_level
        ======== ==========================

    A more thorough description of available sensors and their
    units is available on `Awair's API documentation`_.

    Any new sensors added by an Awair device before this library
    is updated will be accessible via their API name, rather than
    a friendly name.

    .. note::
        The 1st generation Awair device contains an "aggregate dust"
        sensor, which measures a range of particle sizes. It cannot
        distinguish between pm2.5 and pm10 particles.

    .. _`Awair's API documentation`:
      https://docs.developer.getawair.com/?version=latest#data-guide
    """

    def __repr__(self) -> str:
        """Pretty-print."""
        return f"Sensors({super().__repr__()})"



"""Wrapper class for awair airdata responses."""

from datetime import datetime
from typing import Any, Dict

from python_awair import const
from python_awair.indices import Indices
from python_awair.sensors import Sensors


class AirData:
    """Wrapper class for awair airdata responses."""

    timestamp: datetime
    score: float
    sensors: Sensors
    indices: Indices

    def __init__(self, attributes: Dict[str, Any]) -> None:
        """Initialize from API data."""
        self.timestamp = datetime.strptime(attributes["timestamp"], const.DATE_FORMAT)
        self.score = attributes["score"]

        self.sensors = Sensors(
            {
                sensor["comp"]: sensor["value"]
                for sensor in attributes.get("sensors", [])
            }
        )

        self.indices = Indices(
            {index["comp"]: index["value"] for index in attributes.get("indices", [])}
        )

    def __repr__(self) -> str:
        """Return a friendly representation."""
        return (
            f"<AirData@{self.timestamp}: "
            + f"score={round(self.score, 2)} sensors={self.sensors}>"
        )



"""Dict with attribute-like access."""

from typing import Any, Dict, KeysView

from python_awair import const


class AttrDict(Dict[str, Any]):
    """Dict with attribute-like access.

    For example, given an AttrDict *foo*, we could
    access its values via *foo["bar"]* or *foo.bar*.

    This is the parent class for the Sensors and Indices
    classes, and as such it renames some properties to
    friendlier names on initialization (but not anytime after).
    """

    def __init__(self, attrs: Dict[str, Any]) -> None:
        """Initialize, hiding known sensor aliases."""
        new_attrs = dict(attrs)
        for key, value in attrs.items():
            if key in const.SENSOR_TO_ALIAS:
                new_attrs[const.SENSOR_TO_ALIAS[key]] = value
                del new_attrs[key]

        super().__init__(new_attrs)

    def __getattr__(self, name: str) -> Any:
        """Return things in the dict via dot-notation."""
        if name in self:
            return self[name]

        raise AttributeError()

    def __setattr__(self, name: str, value: Any) -> None:
        """Set values in the dict via dot-notation."""
        self[name] = value

    def __delattr__(self, name: str) -> None:
        """Remove values from the dict via dot-notation."""
        del self[name]

    def __dir__(self) -> KeysView[str]:
        """Return dict keys as dir attributes."""
        return self.keys()



"""Class to describe an Awair device."""

import urllib
from abc import ABC, abstractmethod
from datetime import datetime, timedelta
from typing import Any, Dict, List, Optional, Union, cast

import voluptuous as vol

from python_awair import const
from python_awair.air_data import AirData
from python_awair.client import AwairClient

AirDataParam = Union[datetime, bool, int, None]


class AwairBaseDevice(ABC):
    """An Awair device.

    This class serves two purposes - it provides metadata about
    a given Awair device, but it also provides methods that
    retrieve sensor measurements from that device. Available
    information includes things like the model, name, and location
    of a device; and the query interface allows the user to query
    for sensor data in several different samplings, over various
    timeframes.

    .. note::
        While you can instantiate this class by hand (perhaps in
        a test case), you should typically load user devices by
        calling *AwairUser.devices()*.

    Args:
        client: An AwairClient that is used if this AwairDevice
            object needs to query the API.

        attributes: A mapping which describes the device. This
            class expects that the *attributes* provided are
            essentially the result of calling the
            */v1/users/self/devices* API endpoint.
    """

    device_id: int
    """int: The ID that identifies the Awair device."""

    uuid: str
    """str: Another ID that identifies the Awair device.

    This ID typically takes the form of *model_id*. For example,
    given a first-gen Awair device with ID 123, The uuid would be
    "awair_123".
    """

    device_type: str
    """str: The API name for the model of this Awair device.

    This differs from the human-friendly name, which is given
    by the *model* attribute.

    .. table::
        :widths: auto

        =============    ======================
        Device type      Model
        =============    ======================
        awair            `Awair (1st Edition)`_
        awair-element    `Awair Element`_
        awair-glow       `Awair Glow`_
        awair-glow-c     `Awair Glow C`_
        awair-mint       `Awair Baby`_
        awair-omni       `Awair Omni`_
        awair-r2         `Awair (2nd Edition)`_
        =============    ======================

    .. _`Awair (2nd Edition)`: https://getawair.com/pages/awair-2nd-edition
    .. _`Awair Baby`: https://getawair.com/pages/awair-baby
    .. _`Awair Element`: https://getawair.com/pages/awair-element
    .. _`Awair Glow C`: https://getawair.com/pages/awair-glow
    .. _`Awair Glow`:
      https://web.archive.org/web/20161210171139/https://getawair.com/pages/awair-glow
    .. _`Awair Omni`: https://getawair.com/pages/awair-for-business
    .. _`Awair (1st Edition)`:
      https://web.archive.org/web/20150528004143/https://getawair.com/
    """

    mac_address: Optional[str]
    """Optional[str]: The network MAC address."""

    latitude: Optional[float]
    """float: The latitude of the device's location, if known."""

    location_name: Optional[str]
    """float: Description of the device's location."""

    longitude: Optional[float]
    """float: The longitude of the device's location, if known."""

    name: Optional[str]
    """Optional[str]: The user-assigned name for this device."""

    preference: Optional[str]
    """Optional[str]: The device "preference".

    This represents an instruction to the Awair application,
    which represents the area of concern for this device. Put
    differently, it represents *why* the user is using this
    device to monitor air quality - for example, concern about
    allergies.

    Example: "GENERAL"
    """

    room_type: Optional[str]
    """Optional[str]: The type of room this device is in.

    For example, a "LIVING_ROOM" or an "OFFICE".
    """

    space_type: Optional[str]
    """Optional[str]: The type of space this device is in.

    For example, this might be an "OFFICE" or a "HOME".
    """

    timezone: Optional[str]
    """Optional[str]: The timezone of the device."""

    client: AwairClient
    """AwairClient: A reference to the configured AwairClient.

    This is the class that actually queries the API. It's here
    if you need it, but you probably don't need to use it.
    """

    def __init__(self, client: AwairClient, attributes: Dict[str, Any]) -> None:
        """Initialize an awair device from API attributes."""
        self.device_id = attributes["deviceId"]
        self.uuid = attributes["deviceUUID"]
        self.device_type = attributes["deviceType"]
        self.mac_address = attributes.get("macAddress", None)

        self.latitude = attributes.get("latitude", None)
        self.longitude = attributes.get("longitude", None)
        self.name = attributes.get("name", None)
        self.preference = attributes.get("preference", None)
        self.room_type = attributes.get("roomType", None)
        self.space_type = attributes.get("spaceType", None)
        self.timezone = attributes.get("timezone", None)

        self.client = client

    def __repr__(self) -> str:
        """Return a friendly representation."""
        return f"<AwairDevice: uuid={self.uuid} model={self.model}>"

    @property
    def model(self) -> str:
        """Return the human-friendly model, if known.

        .. table::
            :widths: auto

            =============    ======================
            Device type      Model
            =============    ======================
            awair            `Awair (1st Edition)`_
            awair-element    `Awair Element`_
            awair-glow       `Awair Glow`_
            awair-glow-c     `Awair Glow C`_
            awair-mint       `Awair Baby`_
            awair-omni       `Awair Omni`_
            awair-r2         `Awair (2nd Edition)`_
            =============    ======================

        .. _`Awair (2nd Edition)`: https://getawair.com/pages/awair-2nd-edition
        .. _`Awair Baby`: https://getawair.com/pages/awair-baby
        .. _`Awair Element`: https://getawair.com/pages/awair-element
        .. _`Awair Glow C`: https://getawair.com/pages/awair-glow
        .. _`Awair Glow`:
          https://web.archive.org/web/20161210171139/https://getawair.com/pages/awair-glow
        .. _`Awair Omni`: https://getawair.com/pages/awair-for-business
        .. _`Awair (1st Edition)`:
          https://web.archive.org/web/20150528004143/https://getawair.com/
        """
        return const.AWAIR_MODELS.get(self.device_type, self.device_type)

    async def air_data_latest(self, fahrenheit: bool = False) -> Optional[AirData]:
        """Get the latest air data for this device.

        Returns one AirData class describing the most up-to-date measurements
        for this device's sensors. If the device has been offline for more
        than 10 minutes, None will be returned.

        Args:
            fahrenheit: Return temperatures in fahrenheit (the default is to
              return temperatures in celsius). The conversion is done in the
              Awair API itself, not in this library.
        """
        response = await self.__get_airdata("latest", fahrenheit=fahrenheit)
        if response:
            return response[0]

        return None

    async def air_data_five_minute(self, **kwargs: AirDataParam) -> List[AirData]:
        r"""Return five-minute summary air data readings for this device.

        Each data point returned represents a five-minute average of sensor readings.
        Up to a maximum of 288 data points will be returned - which represents 24
        hours of data.

        Args:
            kwargs: A mapping of query parameters, which influence the data returned.
                None are required:

                ==========  =====================================================
                Parameter   Value
                ==========  =====================================================
                fahrenheit  *False* (default): temperature data is returned in
                            celsius.

                            *True*: temperature data is returned in fahrenheit.

                desc        *True* (default): datapoints are ordered descending
                            from the *to* parameter.

                            *False*: datapoints are ordered ascending from the
                            *to* parameter.

                limit       *int*: represents the maximum number of datapoints to
                            return. The default and maximum for this parameter is
                            288.

                from        *datetime*: lower bound for the earliest datapoint to
                            return. May not be chronologically after the *to*
                            parameter, and the difference between the *to* and
                            *from* parameters may not exceed 24 hours.
                            Defaults to 24 hours before the current date/time.

                to          *datetime*: upper bound for the most recent datapoint
                            to return. May not be chronologically before the
                            *from* parameter, and the difference between the *to*
                            and *from* parameters may not exceed 24 hours.
                            Defaults to the current date/time.
                ==========  =====================================================

        """
        return await self.__get_airdata("5-min-avg", **kwargs)

    async def air_data_fifteen_minute(self, **kwargs: AirDataParam) -> List[AirData]:
        r"""Return fifteen-minute summary air data readings for this device.

        Each data point returned represents a fifteen-minute average of sensor readings.
        Up to a maximum of 672 data points will be returned - which represents 7 days
        of data.

        Args:
            kwargs: A mapping of query parameters, which influence the data returned.
                None are required:

                ==========  =====================================================
                Parameter   Value
                ==========  =====================================================
                fahrenheit  *False* (default): temperature data is returned in
                            celsius.

                            *True*: temperature data is returned in fahrenheit.

                desc        *True* (default): datapoints are ordered descending
                            from the *to* parameter.

                            *False*: datapoints are ordered ascending from the
                            *to* parameter.

                limit       *int*: represents the maximum number of datapoints to
                            return. The default and maximum for this parameter is
                            672.

                from        *datetime*: lower bound for the earliest datapoint to
                            return. May not be chronologically after the *to*
                            parameter, and the difference between the *to* and
                            *from* parameters may not exceed 7 days.
                            Defaults to 7 days before the current date/time.

                to          *datetime*: upper bound for the most recent datapoint
                            to return. May not be chronologically before the
                            *from* parameter, and the difference between the *to*
                            and *from* parameters may not exceed 7 days.
                            Defaults to the current date/time.
                ==========  =====================================================

        """
        return await self.__get_airdata("15-min-avg", **kwargs)

    async def air_data_raw(self, **kwargs: AirDataParam) -> List[AirData]:
        r"""Return the raw, per-second air data readings for this device.

        Each data point returned represents the sensor readings at a given second.
        Up to a maximum of 360 data points will be returned - which represents 1 hour
        of data.

        Args:
            kwargs: A mapping of query parameters, which influence the data returned.
                None are required:

                ==========  =====================================================
                Parameter   Value
                ==========  =====================================================
                fahrenheit  *False* (default): temperature data is returned in
                            celsius.

                            *True*: temperature data is returned in fahrenheit.

                desc        *True* (default): datapoints are ordered descending
                            from the *to* parameter.

                            *False*: datapoints are ordered ascending from the
                            *to* parameter.

                limit       *int*: represents the maximum number of datapoints to
                            return. The default and maximum for this parameter is
                            360.

                from        *datetime*: lower bound for the earliest datapoint to
                            return. May not be chronologically after the *to*
                            parameter, and the difference between the *to* and
                            *from* parameters may not exceed 1 hour.
                            Defaults to 1 hour before the current date/time.

                to          *datetime*: upper bound for the most recent datapoint
                            to return. May not be chronologically before the
                            *from* parameter, and the difference between the *to*
                            and *from* parameters may not exceed 1 hour.
                            Defaults to the current date/time.
                ==========  =====================================================

        """
        return await self.__get_airdata("raw", **kwargs)

    @abstractmethod
    def _get_airdata_base_url(self) -> str:
        """Get the base URL to use for airdata."""
        raise TypeError("expected subclass to define override")

    @abstractmethod
    def _extract_airdata(self, response: Any) -> List[Any]:
        """Get the data object out of a response."""
        raise TypeError("expected subclass to define override")

    async def __get_airdata(self, kind: str, **kwargs: AirDataParam) -> List[AirData]:
        """Call one of several varying air-data API endpoints."""
        url = "/".join([self._get_airdata_base_url(), "air-data", kind])

        if kwargs is not None:
            url += self._format_args(kind, **kwargs)

        response = await self.client.query(url)
        return [AirData(data) for data in self._extract_airdata(response)]

    @staticmethod
    def _format_args(kind: str, **kwargs: AirDataParam) -> str:
        max_limit = {"raw": 360, "5-min-avg": 288, "15-min-avg": 672}
        max_hours = {"raw": 1, "15-min-avg": 168, "5-min-avg": 24}

        def validate_hours(params: Dict[str, Any]) -> Dict[str, Any]:
            hour_limit = max_hours.get(kind, 24)
            right_now = datetime.now()
            from_date = params.get("from_date", right_now - timedelta(hours=hour_limit))
            to_date = params.get("to_date", right_now)

            if not hasattr(from_date, "now") or not hasattr(to_date, "now"):
                raise vol.Invalid(
                    "Expected 'from_date' and/or 'to_date' to be instances of datetime"
                )

            if from_date > right_now or to_date > right_now:
                raise vol.Invalid("Dates cannot be in the future!")
            if from_date > to_date:
                raise vol.Invalid("'from_date' cannot be greater than 'to_date'.")
            if (to_date - from_date) > timedelta(hours=hour_limit):
                raise vol.Invalid(
                    "Difference between 'from_date' and 'to_date' must be less than "
                    + f"or equal to {hour_limit} hours."
                )

            if "from_date" in params:
                params["from_date"] = str(params["from_date"])

            if "to_date" in params:
                params["to_date"] = str(params["to_date"])

            return params

        schema = vol.Schema(
            vol.All(
                {
                    vol.Optional("fahrenheit"): vol.All(
                        bool, vol.Coerce(str), vol.Lower
                    ),
                    vol.Optional("desc"): vol.All(bool, vol.Coerce(str), vol.Lower),
                    vol.Optional("limit"): vol.All(
                        int,
                        vol.Range(min=1, max=max_limit.get(kind, 1)),
                        vol.Coerce(str),
                    ),
                    # We validate dates by hand because it's annoying af with mocking.
                    vol.Optional("from_date"): object,
                    vol.Optional("to_date"): object,
                },
                validate_hours,
            )
        )

        args = schema(kwargs)
        if args:
            return "?" + urllib.parse.urlencode(args)

        return ""


class AwairDevice(AwairBaseDevice):
    """A cloud-based Awair device."""

    def _get_airdata_base_url(self) -> str:
        """Get the base URL to use for airdata."""
        return "/".join([const.DEVICE_URL, self.device_type, str(self.device_id)])

    def _extract_airdata(self, response: Any) -> List[Any]:
        """Get the data object out of a response."""
        return cast(List[Any], response.get("data", []))


class AwairLocalDevice(AwairBaseDevice):
    """A local Awair device."""

    device_addr: str
    """The DNS or IP address of the device."""

    def __init__(
        self, client: AwairClient, device_addr: str, attributes: Dict[str, Any]
    ):
        """Initialize an awair local device from API attributes."""
        # the format of the config endpoint for local sensors is different than
        # the cloud API.
        device_uuid: str = attributes["device_uuid"]
        [device_type, device_id_str] = device_uuid.split("_", 1)
        device_id = int(device_id_str)
        attributes["deviceId"] = device_id
        attributes["deviceUUID"] = device_uuid
        attributes["deviceType"] = device_type
        attributes["macAddress"] = attributes.get("wifi_mac", None)
        super().__init__(client, attributes)
        self.device_addr = device_addr

    def _get_airdata_base_url(self) -> str:
        """Get the base URL to use for airdata."""
        return f"http://{self.device_addr}"

    def _extract_airdata(self, response: Any) -> List[Any]:
        """Get the data object out of a response."""
        # reformat local sensors response to match the cloud API
        top_level = {"timestamp", "score"}
        sensors = [
            {"comp": k, "value": response[k]}
            for k in response.keys()
            if k not in top_level
        ]
        data = {
            "timestamp": response["timestamp"],
            "score": response["score"],
            "sensors": sensors,
        }

        return [data]

    @staticmethod
    def _format_args(kind: str, **kwargs: AirDataParam) -> str:
        if "fahrenheit" in kwargs:
            if kwargs["fahrenheit"]:
                raise ValueError("fahrenheit is not supported for local sensors yet")
            # if we pass any URL parameters with local sensors, it causes the
            # timestamp to be the empty string.
            del kwargs["fahrenheit"]

        return AwairBaseDevice._format_args(kind, **kwargs)



"""Wrapper class to query the Awair API."""

from typing import Any, Dict, NoReturn

from aiohttp import ClientResponse, ClientSession

from python_awair.auth import AwairAuth
from python_awair.exceptions import (
    AuthError,
    AwairError,
    NotFoundError,
    QueryError,
    RatelimitError,
)


class AwairClient:
    """Python asyncio client for the Awair GraphQL API."""

    def __init__(self, authenticator: AwairAuth, session: ClientSession,) -> None:
        """Initialize an AwairClient with sensible defaults."""
        self.__authenticator = authenticator
        self.__session = session

    async def query(self, url: str) -> Any:
        """Query the Awair api, and handle errors."""
        headers = await self.__headers()
        async with self.__session.get(url, headers=headers) as resp:
            if resp.status != 200:
                self.__handle_non_200_error(resp)

            json = await resp.json()
            self.__check_errors_array(json)

            return json

    async def __headers(self) -> Dict[str, str]:
        """Return headers to set on the API request."""
        token = await self.__authenticator.get_bearer_token()
        return {
            "Authorization": f"Bearer {token}",
            "Content-Type": "application/json",
        }

    @staticmethod
    def __check_errors_array(json: Dict[Any, Any]) -> None:
        """Check for an "errors" array and process it.

        Holdover from the GraphQL API, unclear if we could still get messages like this.
        """
        if "errors" in json:
            messages = []
            for error in json["errors"]:
                if "Too many requests" in error["message"]:
                    raise RatelimitError()

                messages.append(error.get("message", "Unknown error"))

            if messages:
                raise AwairError(", ".join(messages))

    @staticmethod
    def __handle_non_200_error(resp: ClientResponse) -> NoReturn:
        if resp.status == 400:
            raise QueryError()

        if resp.status == 401 or resp.status == 403:
            raise AuthError()

        if resp.status == 404:
            raise NotFoundError()

        if resp.status == 429:
            raise RatelimitError()

        raise AwairError()



"""Authentication constructs for the Awair API."""

from abc import ABC, abstractmethod


class AwairAuth(ABC):
    """Abstract authentication that provides a Bearer token."""

    @abstractmethod
    async def get_bearer_token(self) -> str:
        """Return a valid bearer token for authentication."""


class AccessTokenAuth(AwairAuth):
    """Authentication that uses an Awair access token."""

    def __init__(self, access_token: str) -> None:
        """Initialize and save off our acces token."""
        self.access_token = access_token
        super().__init__()

    async def get_bearer_token(self) -> str:
        """Return the access token for authentication."""
        return self.access_token



"""Indices dict with attribute access."""
from python_awair.attrdict import AttrDict


class Indices(AttrDict):
    """Indices of an AwairDevice.

    An Indices object represents a set of "index" values for a
    set of sensors of a given AwairDevice. Essentially, the "index"
    is a bit like a quality score - and Awair has devised a set of
    quality levels for a variety of their sensors. The "index" is given
    as a *float* between -4 and 4, but the absolute value is what
    really matters - just ignore the sign. As a value approaches 4.0,
    it is considered "worse". As it approaches 0, it is considered
    "better".

    A mapping of index ranges and values per-sensor can be found
    at Awair's `API documentation`_ - that list is authoritative.

    .. _`API documentation`:
      https://docs.developer.getawair.com/?version=latest#data-guide


    The Indices object is a subclass of AttrDict, and thus its values
    are accessible via string keys - like *foo["bar"]* - or via dot-notation:
    *foo.bar*.

    The index names from the Awair API are not entirely user-friendly,
    so we've aliased known indices to more descriptive values:

    .. table::

        ======== ==========================
        API name python_awair name
        ======== ==========================
        temp     temperature
        humid    humidity
        co2      carbon_dioxide
        voc      volatile_organic_compounds
        pm25     particulate_matter_2_5
        ======== ==========================

    Any new indices added by an Awair device before this library
    is updated will be accessible via their API name, rather than
    a friendly name.

    .. note::
        The 1st generation Awair device will have a "dust" index,
        since it has an aggregate pm2.5/pm10 dust sensor (and cannot
        distinguish between those two sizes).

    .. note::
        Do not assume that every sensor present on a device will
        also have a corresponding "index"; this is not the case.
    """

    def __repr__(self) -> str:
        """Pretty-print."""
        return f"Indices({super().__repr__()})"



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['python_awair']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.6.1,<4.0.0', 'voluptuous>=0.11.7,<0.13.0']

setup_kwargs = {
    'name': 'python-awair',
    'version': '0.2.2',
    'description': 'asyncio client for the Awair GraphQL API',
    'long_description': '# python_awair\n\n![Latest PyPI version](https://img.shields.io/pypi/v/python_awair.svg)\n![CI](https://github.com/ahayworth/python_awair/workflows/CI/badge.svg?branch=master)\n[![codecov](https://codecov.io/gh/ahayworth/python_awair/branch/master/graph/badge.svg)](https://codecov.io/gh/ahayworth/python_awair)\n[![Documentation Status](https://readthedocs.org/projects/python-awair/badge/?version=latest)](https://python-awair.readthedocs.io/en/latest/?badge=latest)\n\nThis is an async library which accesses portions of the [Awair](https://getawair.com) REST API. It exists primarily\nto support the Home Assistant integration, but is considered active and supported by its author. PRs welcome!\n\nFeatures:\n- Object-oriented approach to querying and handling data\n- Supports the "user" portion of the API.\n- Possible to list devices, user information, and to query for a variety of sensor data over various timeframes.\n\nNot yet supported:\n- Device API usage\n- Organization API\n- Device management (such as changing the display of a device)\n\nDive into our [documentation](https://python-awair.readthedocs.io/en/latest) to get started!\n\n# Development\n\n- We manage dependencies and builds via [poetry](https://python-poetry.org)\n- We use [pytest](https://github.com/pytest-dev/pytest) and [tox](https://github.com/tox-dev/tox) to test\n- A variety of linters are available and CI enforces them\n\nAfter installing and configuring poetry:\n- Run `poetry install` to install dev dependencies\n- Run `poetry shell` to drop into a virtualenv\n- Run `poetry run tox` (or just `tox` if you\'re in a virtualenv) to test\n  - Run `poetry run tox -e lint` (or just `tox -e lint` if you\'re in a virtualenv) to run linters.\n',
    'author': 'Andrew Hayworth',
    'author_email': 'ahayworth@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ahayworth/python_awair',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



"""An Awair user."""


from datetime import date
from typing import Any, Dict, List, Optional

from python_awair import const
from python_awair.client import AwairClient
from python_awair.devices import AwairDevice


class AwairUser:
    """An Awair user.

    This class is primarily informational - it describes the
    user to which the provided authentication belongs (be that an
    access_token, or perhaps an OAuth token). Available fields
    include user information, API quotas, and usage. Additionally,
    the class provides a method to access the list of devices
    that are "owned" by this user.

    .. note::
        While you can instantiate this class by hand (perhaps in
        a test case), you should typically load a user by calling
        *Awair.user()* instead.

    Args:
        client: An AwairClient that is used if this AwairUser
            object needs to query the API.

        attributes: A mapping which describes the user. This class
            expects that the *attributes* provided are essentially
            the result of calling the */v1/users/self* API
            endpoint.
    """

    user_id: str
    """str: The user ID uniquely references an Awair user account.
    It is returned as a string, because API docs indicate it is
    a string.
    """

    email: Optional[str]
    """Optional[str]: The email addres on file for the user."""

    first_name: Optional[str]
    """Optional[str]: The first name of the user."""

    last_name: Optional[str]
    """Optional[str]: The last name of the user."""

    sex: Optional[str]
    """Optional[str]: The 'sex' of the user.

    Typical values seem to be 'MALE', 'FEMALE', or 'UNKNOWN'.
    """

    dob: Optional[date]
    """Optional[date]: The user's date of birth, if known."""

    tier: Optional[str]
    """Optional[str]: The account "tier".

    This broadly maps to a set of permissions and API quotas,
    but they are not currently well-defined.
    """

    usages: Dict[str, int]
    """Dict[str, int]: A mapping describing API usage.

    The keys represent the type of API call being described, and
    should be reflected in the user's *permissions* attribute.

    The values represent the number of times that the API call
    has been used in this usage window. Usage windows reset at
    midnight.

    .. todo::
        Document the timezone of the usage window reset.
    """

    permissions: Dict[str, int]
    """Dict[str, int]: A mapping describing API quotas.


    The keys represent the type of API call being described, and
    if the API call has been used in this usage window, then it
    will be represented in this user's *usages* attribute.

    The values represent the maximum number of times that the API
    call may be used in this usage window. Usage windwos reset at
    midnight.

    .. todo::
        Document the timezone of the usage window reset.
    """

    client: AwairClient
    """AwairClient: A reference to the configured AwairClient.

    This AwairClient will be used to query the API and validate
    HTTP responses. It's normally not something one would need to
    access.
    """

    def __init__(self, client: AwairClient, attributes: Dict[str, Any]) -> None:
        """Initialize an awair user from API attributes."""
        self.user_id = attributes["id"]
        self.email = attributes.get("email", None)
        self.first_name = attributes.get("firstName", None)
        self.last_name = attributes.get("lastName", None)

        self.sex = attributes.get("sex", None)
        self.dob: Optional[date]
        dob_day = attributes.get("dobDay", None)
        dob_month = attributes.get("dobMonth", None)
        dob_year = attributes.get("dobYear", None)

        if all([dob_day, dob_month, dob_year]):
            self.dob = date(day=dob_day, month=dob_month, year=dob_year)
        else:
            self.dob = None

        self.tier = attributes.get("tier", None)
        self.usages = {
            item["scope"]: item["usage"] for item in attributes.get("usages", [])
        }
        self.permissions = {
            item["scope"]: item["quota"] for item in attributes.get("permissions", [])
        }

        self.client = client

    def __repr__(self) -> str:
        """Return a friendly representation."""
        parts = [f"user_id={self.user_id}"]
        if self.email is not None:
            parts.append(f"email={self.email}")

        return f"<AwairUser: {' '.join(parts)}>"

    async def devices(self) -> List[AwairDevice]:
        """Return a list of awair devices this user owns.

        .. note::
            For organization users, the underlying API endpoint
            that this method uses should show you devices that
            you have access to view. However, the organization
            API is not directly supported at this time.
        """
        response = await self.client.query(const.DEVICE_URL)
        return [
            AwairDevice(client=self.client, attributes=device)
            for device in response.get("devices", [])
        ]



"""Python asyncio client for the Awair REST API.

This module is an object-oriented wrapper around the Awair_
REST API_. It requires an access token (which can be obtained
from the `developer console`_) and implements read-only access
to the "user" portions of the API.

.. _Awair: https://getawair.com
.. _API: https://docs.developer.getawair.com/?version=latest
.. _`developer console`: https://developer.getawair.com
"""

from asyncio import gather
from typing import List, Optional

from aiohttp import ClientSession

from python_awair import const
from python_awair.auth import AccessTokenAuth, AwairAuth
from python_awair.client import AwairClient
from python_awair.devices import AwairLocalDevice
from python_awair.exceptions import AwairError
from python_awair.user import AwairUser


class Awair:
    """Entry class for the Awair API.

    Args:
        session: An aiohttp session that will be used to query
            the Awair API.
        access_token: An optional access token, obtained from
            the Awair developer console, used to authenticate
            to the Awair API.
        authenticator: An optional instance of an AwairAuth class,
            which can provide an HTTP Bearer token for
            authentication. Most users will simply provide an
            access_token, instead.
    """

    client: AwairClient
    """AwairClient: The instantiated AwairClient
        that will be used to fetch API responses and
        check for HTTP errors.
    """

    def __init__(
        self,
        session: ClientSession,
        access_token: Optional[str] = None,
        authenticator: Optional[AwairAuth] = None,
    ) -> None:
        """Initialize the Awair API wrapper."""
        if authenticator:
            self.client = AwairClient(authenticator, session)
        elif access_token:
            self.client = AwairClient(AccessTokenAuth(access_token), session)
        else:
            raise AwairError("No authentication supplied!")

    async def user(self) -> AwairUser:
        """Return the current AwairUser.

        The Awair "user" API does not provide a way to query for
        a specific user, so this method always returns the user
        that is associated with the authentication that is in-use.
        This is *typically* the user that owns the access_token
        that was provided at class instantiaton, unless you have
        provided an authenticator class implementing some other
        stategy (perhaps OAuth).
        """
        response = await self.client.query(const.USER_URL)
        return AwairUser(client=self.client, attributes=response)


class AwairLocal:
    """Entry class for the local sensors Awair API."""

    client: AwairClient
    """AwairClient: The instantiated AwairClient
        that will be used to fetch API responses and
        check for HTTP errors.
    """

    _device_addrs: List[str]
    """IP or DNS addresses of Awair devices with the local sensors API enabled."""

    def __init__(self, session: ClientSession, device_addrs: List[str]) -> None:
        """Initialize the Awair local sensors API wrapper."""
        self._device_addrs = device_addrs
        if len(device_addrs) > 0:
            self.client = AwairClient(AccessTokenAuth(""), session)
        else:
            raise AwairError("No local Awair device addresses supplied!")

    async def devices(self) -> List[AwairLocalDevice]:
        """Return a list of local awair devices."""
        responses = await gather(
            *(
                self.client.query(f"http://{addr}/settings/config/data")
                for addr in self._device_addrs
            )
        )
        return [
            AwairLocalDevice(
                client=self.client, device_addr=self._device_addrs[i], attributes=device
            )
            for i, device in enumerate(responses)
        ]



"""Mostly query constants."""
BASE_URL = "https://developer-apis.awair.is/v1"
USER_URL = f"{BASE_URL}/users/self"
DEVICE_URL = f"{USER_URL}/devices"
DATE_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"

SENSOR_TO_ALIAS = {
    "temp": "temperature",
    "humid": "humidity",
    "co2": "carbon_dioxide",
    "voc": "volatile_organic_compounds",
    "pm25": "particulate_matter_2_5",
    "lux": "illuminance",
    "spl_a": "sound_pressure_level",
}

AWAIR_MODELS = {
    "awair": "Awair",
    "awair-element": "Awair Element",
    "awair-glow": "Awair Glow",
    "awair-glow-c": "Awair Glow C",
    "awair-mint": "Awair Mint",
    "awair-omni": "Awair Omni",
    "awair-r2": "Awair 2nd Edition",
}
